#include <iostream>
#include <conio.h>
#include <windows.h>

using namespace std;

// A simple game object system where each object type has the same update function
// To update the system, a separate loop is used for each type
namespace Simple
{
    class Fighter
    {
    public:
        void update( float dt )
        {
        }
        double a,b,c;
    };
    class Bomber
    {
    public:
        void update( float dt )
        {
        }
        double a,b,c;
    };
    class Recon
    {
    public:
        void update( float dt )
        {
        }
        double a,b,c;
    };
    class System
    {
    public:
        System()
        {
            for (int i=0; i<5000; i++)
            {
                fighters[i] = new Fighter();
                bombers[i] = new Bomber();
                recons[i] = new Recon();
            }
        }
        ~System()
        {
            for (int i=0; i<5000; i++)
            {
                delete fighters[i];
                delete bombers[i];
                delete recons[i];
            }
        }
        Fighter *fighters[5000];
        Bomber *bombers[5000];
        Recon *recons[5000];
    };
}

//The same game object system that uses the template pattern
// and polymorhpism to update all objects with only one loop
namespace Polymorphism
{
    class GameObject
    {
    public:
        virtual void update( float dt ) = 0;
        double a,b,c;
    };
    class Fighter : public GameObject
    {
    public:
        void update( float dt )
        {
        }
    };
    class Bomber : public GameObject
    {
    public:
        void update( float dt )
        {
        }
    };
    class Recon : public GameObject
    {
    public:
        void update( float dt )
        {
        }
    };
    class System
    {
    public:
        System()
        {
            //create objects out of order
            for (int i=0; i<5000; i++)
            {
                fighters[i] = new Fighter();
                bombers[i] = new Bomber();
                recons[i] = new Recon();
                if ( i % 3 == 0 )
                {
                    objects[i] = fighters[i];
                    objects[i+5000] = bombers[i];
                    objects[i+10000] = recons[i];
                }
                else
                if ( i % 3 == 1 )
                {
                    objects[i] = bombers[i];
                    objects[i+5000] = recons[i];
                    objects[i+10000] = fighters[i];
                }
                else
                if ( i % 3 == 2 )
                {
                    objects[i] = recons[i];
                    objects[i+5000] = fighters[i];
                    objects[i+10000] = bombers[i];
                }
            }
        }
        ~System()
        {
            for (int i=0; i < 15000; i++)
            {
                delete objects[i];
            }
        }
        GameObject *objects[15000];
        Fighter *fighters[5000];
        Bomber *bombers[5000];
        Recon *recons[5000];
    };
}

//Another version of the game object system, where each object type
// is managed by a bucket class, so only 3 virtual function calls are
// needed
namespace Buckets
{
    class GameObject
    {
    public:
        virtual void update( float dt ) = 0;
        double a,b,c;
    };
    class Fighter : public GameObject
    {
    public:
        void update( float dt )
        {
        }
    };
    class Bomber : public GameObject
    {
    public:
        void update( float dt )
        {
        }
    };
    class Recon : public GameObject
    {
    public:
        void update( float dt )
        {
        }
    };

    class Bucket
    {
    public:
        virtual void updateAll( float dt ) = 0;
    };

    template<class T>
    class ConcreteBucket : public Bucket
    {
    public:
        ConcreteBucket()
        {
			objects = new T[5000];
        }
        ~ConcreteBucket()
        {
			delete [] objects;
        }
        void updateAll( float dt )
        {
            for (int i=0; i<5000; i++)
            {
                objects[i].update( dt );
            }
        };

        T *objects;
    };

    class System
    {
    public:
        System()
        {
            buckets[0] = &fighters;
            buckets[1] = &bombers;
            buckets[2] = &recons;
        }
        ~System()
        {
        }
        ConcreteBucket<Buckets::Fighter> fighters;
        ConcreteBucket<Buckets::Bomber> bombers;
        ConcreteBucket<Buckets::Recon> recons;

        Bucket *buckets[3];
    };
}


void runSimple( Simple::System &simpleSys )
{
    __int64 freq, tStart, tStop;
    unsigned long total=0;
    double ave;

    QueryPerformanceFrequency((LARGE_INTEGER*)&freq);

    cout << "Timing simple system..." << endl;
    for ( int t = 0; t < 100; t++)
    {
        QueryPerformanceCounter((LARGE_INTEGER*)&tStart);
        for (int i=0; i<5000; i++)
        {
            simpleSys.fighters[i]->update( 0 );
        }
        for (int i=0; i<5000; i++)
        {
            simpleSys.bombers[i]->update( 0 );
        }
        for (int i=0; i<5000; i++)
        {
            simpleSys.recons[i]->update( 0 );
        }
        QueryPerformanceCounter((LARGE_INTEGER*)&tStop);
        total += (unsigned long)(((tStop - tStart) *1000000) / freq);
    }
    ave = total / 100000.0;
    cout << "simple version ran in average " << ave << " seconds" << endl;
}
void runPoly( Polymorphism::System &polymorphicSys )
{
    __int64 freq, tStart, tStop;
    unsigned long total=0;
    double ave;

    QueryPerformanceFrequency((LARGE_INTEGER*)&freq);

    cout << "Timing polymorphic system..." << endl;
    for ( int t = 0; t < 100; t++)
    {
        QueryPerformanceCounter((LARGE_INTEGER*)&tStart);
        for(int i=0; i < 15000; i++)
        {
            polymorphicSys.objects[i]->update( 0 );
        }
        QueryPerformanceCounter((LARGE_INTEGER*)&tStop);
        total += (unsigned long)(((tStop - tStart) *1000000) / freq);
    }
    ave = total / 100000.0;
    cout << "polymorphic version ran in average " << ave << " seconds" << endl;

}
void runBucket( Buckets::System &bucketSys )
{
    __int64 freq, tStart, tStop;
    unsigned long total=0;
    double ave;

    QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
    
    cout << "Timing bucket system..." << endl;
    for ( int t = 0; t < 100; t++)
    {
        QueryPerformanceCounter((LARGE_INTEGER*)&tStart);
        for ( int i=0; i< 3; i++ )
        {
            bucketSys.buckets[i]->updateAll( 0 );
        }
        QueryPerformanceCounter((LARGE_INTEGER*)&tStop);
        total += (unsigned long)(((tStop - tStart) *1000000) / freq);
    }
    ave = total / 100000.0;
    cout << "bucket version ran in average " << ave << " seconds" << endl;
}

int main(int argc, const char* argv[])
{
    Buckets::System bucketSys;
    Simple::System simpleSys;
    Polymorphism::System polymorphicSys;

    runSimple( simpleSys );
    runPoly( polymorphicSys );
    runBucket( bucketSys );

	cout << endl << "Press q to quit" << endl;
    char ch = 0;
    do
    {
        ch = _getch();

    }while( ch != 'q');

	return 0;
}
